/* Copyright 2000 John Fremlin
   Distributed under GPL, see file COPYING
*/
/* This code is NOT THREAD SAFE (it uses strerror) */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <signal.h>

#include "config.h"

#ifndef USE_UTMP
#define USE_UTMP 1
#endif

#include <errno.h>
#include <unistd.h>

#ifdef USE_UTMP
#include <utmp.h>
#endif

#ifdef HAVE_CURSES_H
#include <curses.h>
#endif
#ifdef HAVE_NCURSES_H
#include <ncurses.h>
#endif
#ifdef HAVE_TERM_H
#include <term.h>
#endif

#include "system.h"

#ifndef LOGIN_MESSAGE
#define LOGIN_MESSAGE gt("%s login: ")
#endif

inline static
void
clear_tty()
{
#ifdef HAVE_CURSES
  int error;

  if(setupterm(0,STDOUT_FILENO,&error)==ERR){
    fprintf(stderr,"%s: can't setupterm\n",get_program_invocation_name());
    return;
  }
  if(tputs(clear_screen,1,putchar)==ERR){
    fprintf(stderr,"%s: can't clear screen\n",get_program_invocation_name());
    return;
  }
#endif  
}

inline
static
void
utmp_admin(const char* ttyname,const char*hostname)
{
#if USE_UTMP
  struct utmp* utp;
  struct utmp entry;
  char const*tty_basename=ttyname;
  char const*tty_number=0;
  pid_t pid;
  
  if(utmpname(_PATH_UTMP)==-1)
    {
      perror("utmpname");
      return;
    }

  if(ttyname)
  {
    char const*p;
    char const*last_slash=0;
    for(p=ttyname;*p;p++)
	if(*p=='/')last_slash=p;

    if(last_slash)tty_basename = last_slash+1;
    while(--p!=ttyname)
      if(!isdigit(*p))
	{
	  tty_number=p+1;
	  break;
	}
  }
  
  setutent();
  
  pid=getpid();
  while((utp=getutent()))
    if(utp->ut_pid==pid)
      {
	memcpy(&entry,utp,sizeof(entry));
	goto got_entry;
      }

  memset(&entry,0,sizeof(entry));
  strncpy(entry.ut_id, tty_number, sizeof(entry.ut_id));
  /* Hmm. Is this a good thing to do? agetty does it, but perhaps
     leaving a value of zero would make more sense. */
  
 got_entry:  

  strncpy(entry.ut_user, "LOGIN", sizeof(entry.ut_user));
  strncpy(entry.ut_line, tty_basename, sizeof(entry.ut_line));
  strncpy(entry.ut_host, hostname, sizeof(entry.ut_host));
  
  {
    entry.ut_time = time(0);
    entry.ut_type = LOGIN_PROCESS;
    entry.ut_pid = pid;
    entry.ut_session = pid;
  }
  
  if(!pututline(&entry))
      perror("pututline");

  endutent();
      
  updwtmp(_PATH_WTMP,&entry);

#endif
}

void
fill_hostname(char**hostname)
{
  unsigned size = 4;
  /*leave some space for allocator to waste in a 1024 block*/

  while(1)
    {
      *hostname = (char*)malloc(size);
      if(!*hostname)
	{
	  fprintf(stderr,gt("%s: not enough memory to allocate hostname\n"),
		  get_program_invocation_name())
	    ;
	  return;
	}
      
      if(gethostname(*hostname,size))
	{
	  if(errno==ENAMETOOLONG)
	    {
	      size*=2;
	      free(*hostname);
	      continue;
	    }
	  else
	    {
	      fprintf(stderr,gt("%s: unable to determine hostname: %s\n"),
		      get_program_invocation_name(),strerror(errno));
	      free(*hostname);
	      *hostname=0;
	      return;
	    }
	}
      return;
    }

}


int
display(const char*file)
{
  FILE* stream = fopen(file,"r");
  int c;
  
  if(!stream)
    return 0;

  while((c=fgetc(stream))!=EOF)
    fputc(c,stdout);

  fclose(stream);
  
  return 0;
}

int print_prompt(const char*hostname)
{
  char const*host;
  host = hostname ? hostname : gt("[unknown]");
  
  printf(LOGIN_MESSAGE,host);
  fflush(stdout);
  
  return 0;
}

int read_username(void)
{
  char username[UT_NAMESIZE+1];
  char*ptr;
  
  if(!fgets(username,sizeof(username),stdin))
    {
      fprintf(stderr,gt("%s: problem reading username\n"),
	      get_program_invocation_name());
      return 1;
    }
  for(ptr=username;*ptr;ptr++)
    {
      if(*ptr=='\n')
	{
	  *ptr=0;
	  goto polished_off_name;
	}
    }
  fprintf(stderr,gt("%s: username too long\n"),
	      get_program_invocation_name());
  return 2;
  
 polished_off_name:
  if(execl(LOGIN_PROGRAM,LOGIN_PROGRAM,username,0))

  if(execl(LOGIN_PROGRAM,LOGIN_PROGRAM,0))
    {
      fprintf(stderr,gt("%s: can't exec %s: %s\n"),
	      get_program_invocation_name(),
	      LOGIN_PROGRAM,strerror(errno));
      
      return 3;
    }
  return 0;//keep compiler happy
}

int
main()
{
  char*hostname=0;

  gettext_init();

#if 0  
  /* Stop user killing us (is there any security issue?) */
  signal(SIGINT,SIG_IGN);
  signal(SIGIOT,SIG_IGN);
  signal(SIGQUIT,SIG_IGN);
#endif

  /* shouldn't give it the real hostname unless it has actually been
     specified */
  utmp_admin(ttyname(0),hostname?hostname:"");

  clear_tty();
  
  if(display(ETC_ISSUE))return 1;

  fill_hostname(&hostname);
  
  if(print_prompt(hostname))return 2;

  if(hostname)
    free(hostname);
  
  if(read_username())return 3;
  
  return 4;
}
